#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.


import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from xenonpy.utils import TimedMetaClass
from collections import defaultdict
from collections.abc import Iterable


class LogLikelihoodError(Exception):
    """Base exception for LogLikelihood classes"""
    pass


class ResampleError(Exception):
    """Base exception for Resample classes"""
    pass


class ProposalError(Exception):
    """Base exception for Proposal classes"""
    old_smi = None


class SMCError(Exception):
    """Base exception for SMC classes"""
    pass


class BaseLogLikelihood(BaseEstimator, metaclass=TimedMetaClass):

    def fit(self, X, y, **kwargs):
        return self

    def __call__(self, X, **targets):
        return self.log_likelihood(X, **targets)

    def log_likelihood(self, X, **targets):
        """
        Log likelihood

        Parameters
        ----------
        X: list[object]
            Input samples for likelihood calculation.
        targets: tuple[float, float]
            Target area.
            Should be a tuple which have down and up boundary.
            e.g: ``target1=(10, 20)`` equal to ``target1 should in range [10, 20]``.

        Returns
        -------
        log_likelihood: pd.Dataframe of float (col - properties, row - samples)
            Estimated log-likelihood of each sample's property values.
            Cannot be pd.Series!
        """
        # raise NotImplementedError('<log_likelihood> have no implementation')
        raise NotImplementedError('<log_likelihood> method must be implemented')


class BaseLogLikelihoodSet(BaseEstimator, metaclass=TimedMetaClass):
    
    """
    Abstract class to organize log-likelihoods.
    
    Examples
    --------
    .. code::
        class MyLogLikelihood(BaseLogLikelihoodSet):
            def __init__(self):
                super().__init__()
    
                self.loglike1 = SomeFeature1()
                self.loglike1 = SomeFeature2()
                self.loglike2 = SomeFeature3()
                self.loglike2 = SomeFeature4()
    """
    
    def __init__(self, *, loglikelihoods='all'):
        """
            
        Parameters
        ----------
        loglikelihoods: list[str] or 'all'
            log-likelihoods that will be used.
            Default is 'all'.
        """
        self.loglikelihoods = loglikelihoods
        self.__loglikelihoods__ = set()
        self.__loglikelihood_sets__ = defaultdict(list)
    
    @property
    def elapsed(self):
        return self._timer.elapsed
        
    def __setattr__(self, key, value):
        
        if key == '__loglikelihood_sets__':
            if not isinstance(value, defaultdict):
                raise RuntimeError('Can not set "self.__loglikelihood_sets__" by yourself')
            super().__setattr__(key, value)
        if isinstance(value, BaseLogLikelihood):
#            if value.__class__.__name__ in self.__loglikelihoods__:
#                raise RuntimeError('Duplicated log-likelihood <%s>' % value.__class__.__name__)
            self.__loglikelihood_sets__[key].append(value)
            self.__loglikelihoods__.add(value.__class__.__name__)
        else:
            super().__setattr__(key, value)

    @property
    def all_loglikelihoods(self):
        return list(self.__loglikelihoods__)
    
    def _check_input(self, X, y=None, **kwargs):
        def _reformat(x):
            if x is None:
                return x
            
            keys = list(self.__loglikelihood_sets__.keys())
            if len(keys) == 1:
                if isinstance(x, list):
                    return pd.DataFrame(pd.Series(x), columns=keys)
                
                if isinstance(x, np.ndarray):
                    if len(x.shape) == 1:
                        return pd.DataFrame(x, columns=keys)
            
                if isinstance(x, pd.Series):
                    return pd.DataFrame(x.values, columns=keys, index=x.index)
        
            if isinstance(x, pd.Series):
                x = pd.DataFrame(x)
                    
            if isinstance(x, pd.DataFrame):
                tmp = set(x.columns) | set(kwargs.keys())
                if set(keys).isdisjoint(tmp):
                    raise KeyError(
                        'name of columns do not match any log-likelihood set')
                return x
                
            raise TypeError(
                'you can not ues a array-like input'
                'because there are multiply log-likelihood sets or the dim of input is not 1')
            
        return _reformat(X), _reformat(y)
            
    def __call__(self, X, **kwargs):
        return self.log_likelihood(X, **kwargs)

    def log_likelihood(self, X, **kwargs):
        """
        Log likelihood
        
        Parameters
        ----------
        X: list[object]
            Input samples for likelihood calculation.
        kwargs: list[string]
            specified BaseLogLikelihood.
        
        Returns
        -------
        log_likelihood: pd.Dataframe of float (col - properties, row - samples)
            Estimated log-likelihood of each sample's property values.
        """
        # raise NotImplementedError('<log_likelihood> have no implementation')
        #raise NotImplementedError('<log_likelihood> method must be implemented')
            
        if not isinstance(X, Iterable):
            raise TypeError('parameter "entries" must be a iterable object')
                
        if len(X) is 0:
            return None
                        
        if 'loglikelihoods' in kwargs:
            loglikelihoods = kwargs['loglikelihoods']
            if loglikelihoods != 'all' and not isinstance(loglikelihoods, list):
                raise TypeError('parameter "loglikelihoods" must be a list')
        else:
            loglikelihoods = self.loglikelihoods
                            
        #if 'return_type' in kwargs:
        #    del kwargs['return_type']

        results = []

        X, _ = self._check_input(X, **kwargs)
        for k, lls in self.__loglikelihood_sets__.items():
            if k in kwargs:
                k = kwargs[k] # what is this for? even if k not in kwargs... nothing happen right?
            if k in X:
                for f in lls:
                    if loglikelihoods != 'all' and f.__class__.__name__ not in loglikelihoods:
                        continue
                    ret = f.log_likelihood(X[k])
                    results.append(ret)
                                
        return pd.concat(results, axis=1)


class BaseResample(BaseEstimator, metaclass=TimedMetaClass):

    def fit(self, X, y=None, **kwargs):
        return self

    def __call__(self, X, size, p):
        return self.resample(X, size, p)

    def resample(self, X, size, p):
        """
        Re-sample from given samples.

        Parameters
        ----------
        X: list of object
            Input samples for likelihood calculation.
        size: int
            Resample size.
        p: np.ndarray of float
            The probabilities associated with each entry in X.
            If not given the sample assumes a uniform distribution over all entries.

        Returns
        -------
        new_sample: list of object
            Re-sampling result.
        """
        raise NotImplementedError('<resample> method must be implemented')


class BaseProposal(BaseEstimator, metaclass=TimedMetaClass):
    def fit(self, X, y, **kwargs):
        return self

    def __call__(self, X):
        return self.proposal(X)

    def on_errors(self, error):
        raise error

    def proposal(self, X):
        """
        Proposal new samples based on the input samples.

        Parameters
        ----------
        X: list of object
            Samples for generate next samples

        Returns
        -------
        samples: list of object
            Generated samples from input samples.
        """
        raise NotImplementedError('<proposal> method must be implemented')


class BaseSMC(BaseEstimator, metaclass=TimedMetaClass):
    _log_likelihood = None
    _proposal = None
    _resample = None

    # _targets = None

    def log_likelihood(self, X):
        """
        Log likelihood

        Parameters
        ----------
        X: list of object
            Input samples for likelihood calculation.

        Returns
        -------
        log_likelihood: np.ndarray of float
            Estimated likelihood of each samples.
        """
        if self._log_likelihood is None:
            raise NotImplementedError('user need to implement <log_likelihood> method or'
                                      'set <self._log_likelihood> to a instance of <BaseLogLikelihood>')
        # return self._log_likelihood(X, **self._targets)
        return self._log_likelihood(X)

    def resample(self, X, size, p):
        """
        Re-sample from given samples.

        Parameters
        ----------
        X: list[object]
            Input samples for likelihood calculation.
        size: int
            Resample size.
        p: numpy.ndarray[float]
            The probabilities associated with each entry in X.
            If not given the sample assumes a uniform distribution over all entries.

        Returns
        -------
        re-sample: list of object
            Re-sampling result.
        """
        if self._resample is None:
            raise NotImplementedError('user need to implement <resample> method or'
                                      'set <self._resample> to a instance of <BaseResample>')
        return self._resample(X, size, p)

    def proposal(self, X):
        """
        Proposal new samples based on the input samples.

        Parameters
        ----------
        X: list of object
            Samples for generate next samples

        Returns
        -------
        samples: list of object
            Generated samples from input samples.
        """
        if self._proposal is None:
            raise NotImplementedError('user need to implement <proposal> method or'
                                      'set <self._proposal> to a instance of <BaseProposal>')
        return self._proposal(X)

    def on_errors(self, ite, samples, error):
        raise error

    def unique(self, X):
        """

        Parameters
        ----------
        X: list of object
            Input samples.

        Returns
        -------
        unique: list of object
            The sorted unique values.
        unique_counts: np.ndarray of int
            The number of times each of the unique values comes up in the original array
        """
        return np.unique(X, return_counts=True)

    def __setattr__(self, key, value):
        if key is '_log_likelihood' and not isinstance(value, (BaseLogLikelihood, BaseLogLikelihoodSet)):
            raise TypeError(
                '<self._log_likelihood> must be a subClass of <BaseLogLikelihood> or <BaseLogLikelihoodSet>')
        if key is '_proposal' and not isinstance(value, BaseProposal):
            raise TypeError('<self._proposal> must be a subClass of <BaseProposal>')
        if key is '_resample' and not isinstance(value, BaseResample):
            raise TypeError('<self._resample> must be a subClass of <BaseResample>')
        object.__setattr__(self, key, value)

    def __call__(self, samples, beta, *, size=None, yield_lpf=False):
        """
        Run SMC

        Parameters
        ----------
        samples: list of object
            Initial samples.
        beta: list/1D-numpy of float or pd.Dataframe
            Annealing parameters for each step.
            If pd.Dataframe, column names should follow keys of mdl in BaseLogLikeihood or BaseLogLikelihoodSet
        size: int
            Sample size for each draw.
        yield_lpf : bool
            Yield estimated log likelihood, probability and frequency of each samples. Default is ``False``.

        Yields
        -------
        samples: list of object
            New samples in each SMC iteration.
        llh: np.ndarray float
            Estimated values of log-likelihood of each samples.
            Only yield when ``yield_lpf=Ture``.
        p: np.ndarray of float
            Estimated probabilities of each samples.
            Only yield when ``yield_lpf=Ture``.
        freq: np.ndarray of float
            The number of unique samples in original samples.
            Only yield when ``yield_lpf=Ture``.
        """

        # sample size will be set to the length of init_samples if None
        if size is None:
            size = len(samples)

        #         if targets:
        #             self.update_targets(**targets)
        #         else:
        #             if not self._targets:
        #                 raise ValueError('must set targets area')

        try:
            unique, frequency = self.unique(samples)
            ll = self.log_likelihood(unique)

            if isinstance(beta, pd.DataFrame):
                beta = beta[ll.columns].values
            else:
                # assume only one row for beta (1-D list or numpy vector)
                beta = np.transpose(np.repeat([beta], ll.shape[1], axis=0))

            w = np.dot(ll.values, beta[0]) + np.log(frequency)
            w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
            p = np.exp(w - w_sum)
            if yield_lpf:
                yield unique, ll, p, frequency
            else:
                yield unique

            # beta = np.delete(beta,0,0) #remove first "row"

        except SMCError as e:
            self.on_errors(0, samples, e)
        except Exception as e:
            raise e

        # translate between input representation and execute environment representation.
        # make sure beta is not changed (np.delete will return the deleted version, without changing original vector)
        for i, step in enumerate(np.delete(beta, 0, 0)):
            try:
                re_samples = self.resample(unique, size, p)
                samples = self.proposal(re_samples)

                unique, frequency = self.unique(samples)

                # annealed likelihood in log - adjust with copy counts
                ll = self.log_likelihood(unique)
                w = np.dot(ll.values, step) + np.log(frequency)
                w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
                p = np.exp(w - w_sum)
                if yield_lpf:
                    yield unique, ll, p, frequency
                else:
                    yield unique

            except SMCError as e:
                self.on_errors(i + 1, samples, e)
            except Exception as e:
                raise e



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.
import types
from datetime import timedelta
from functools import wraps

import torch
import torch.nn.functional as F
from sklearn.base import RegressorMixin

from .base import BaseRunner


def persist(*args, **kwargs):
    """
    Parameters
    ----------
    args
    kwargs
    Returns
    -------
    """
    n_args = len(args)
    n_kwargs = len(kwargs)

    def _checked(o):
        if not isinstance(o, BaseRunner):
            raise TypeError('persistence only decorate <BaseRunner> inherent object\'s method')
        return o

    # only args or kwargs, no mix
    if (n_args == 0 and n_kwargs == 0) or (n_args != 0 and n_kwargs != 0):
        raise RuntimeError('Decorator need a function or method as first para')

    # if first para is func
    if n_args == 1:
        if isinstance(args[0], (types.FunctionType, types.MethodType)):
            fn = args[0]

            @wraps(fn)
            def _func(self, *args_, **kwargs_):
                self = _checked(self)
                ret = fn(self, *args_, **kwargs_)
                checker = getattr(self._checker, fn.__name__)
                checker.save(ret)
                return ret

            return _func

    # for name paras
    if n_args >= 1:
        if not all([isinstance(o, str) for o in args]):
            raise TypeError('Name of key must be str')

        def _deco(fn_):

            @wraps(fn_)
            def _func_1(self, *args_, **kwargs_):
                self = _checked(self)
                ret = fn_(self, *args_, **kwargs_)
                if not isinstance(ret, tuple):
                    ret = (ret,)
                _n_ret = len(ret)
                if _n_ret != n_args:
                    raise RuntimeError('Number of keys not equal values\' number')
                pair = zip(args, ret)
                checker = getattr(self._checker, fn_.__name__)
                checker.save(**{k: v for k, v in pair})
                return ret if len(ret) > 1 else ret[0]

            return _func_1

        return _deco

    if n_kwargs >= 1:
        if not all([isinstance(v, type) for v in kwargs.values()]):
            raise RuntimeError('Values must be type')

        def _deco(fn_):

            @wraps(fn_)
            def _func_2(self, *args_, **kwargs_):
                self = _checked(self)
                ret = fn_(self, *args_, **kwargs_)
                if not isinstance(ret, tuple):
                    ret = (ret,)
                _n_ret = len(ret)
                if _n_ret != n_kwargs:
                    raise RuntimeError('Number of keys not equal values\' number')
                types_ = kwargs.values()
                if not all([isinstance(v, t) for v, t in zip(ret, types_)]):
                    raise TypeError('Returns\' type not match')
                names = kwargs.keys()
                pair = zip(names, ret)
                checker = getattr(self._checker, fn_.__name__)
                checker.save(**{k: v for k, v in pair})
                return ret if len(ret) > 1 else ret[0]

            return _func_2

        return _deco


class RegressionRunner(BaseRunner, RegressorMixin):
    """
    Run model.
    """

    def __init__(self,
                 epochs=2000,
                 *,
                 cuda=False,
                 check_step=100,
                 log_step=0,
                 work_dir=None,
                 verbose=True,
                 describe=None):
        """
        Parameters
        ----------
        epochs: int
        log_step: int
        cuda: str
        check_step: int
        work_dir: str
        verbose: bool
            Print :class:`ModelRunner` environment.
        """
        super(RegressionRunner, self).__init__(
            epochs, cuda=cuda, work_dir=work_dir, verbose=verbose, describe=describe)
        self._check_step = check_step
        self._log_step = log_step
        self._lr = 0.01
        self._lr_scheduler = None

        self.logger('Runner environment:', 'Running dir: {}'.format(self._work_dir),
                    'Epochs: {}'.format(self._epochs), 'Context: {}'.format(self._device),
                    'Check step: {}'.format(self._check_step), 'Log step: {}\n'.format(
                self._log_step))

    @property
    def lr(self):
        return self._lr

    @lr.setter
    def lr(self, v):
        """"""
        self._lr = v

    @property
    def lr_scheduler(self):
        return None

    @lr_scheduler.setter
    def lr_scheduler(self, v):
        self._lr_scheduler = v

    # @persist('y_true', 'y_pred')
    def post_predict(self, y_true, y_pred):
        y_pred = y_pred.cpu().detach().numpy()
        if y_true is not None:
            return y_true, y_pred
        return y_pred

    def optim(self, iter_):
        # optimization
        optim = torch.optim.Adam(self._model.parameters(), lr=self._lr)

        # adjust learning rate
        scheduler = self._lr_scheduler(optim) if self._lr_scheduler else None
        y_train = y_pred = loss = None
        start = self.elapsed
        for y_train, y_pred, t, i in iter_:

            if scheduler and not isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                scheduler.step()
            loss = F.mse_loss(y_pred, y_train)
            if scheduler and isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                scheduler.step(loss)
            optim.zero_grad()
            loss.backward()
            optim.step()

            if self._log_step > 0 and t % self._log_step == 0 and i == 1:
                elapsed = str(timedelta(seconds=self.elapsed - start))
                start = self.elapsed
                self.logger('{}/{}, Loss={:.4f}, elapsed time: {}'.format(
                    t, self._epochs, loss, elapsed))
            if self._check_step > 0 and t % self._check_step == 0 and i == 1:
                self.checkpoint(mse_loss=loss.item())

        self.logger('Final loss={:.4f}'.format(loss), '')

        return y_train.cpu().detach().numpy(), y_pred.cpu().detach().numpy()



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

__all__ = ['IQSPR', 'GaussianLogLikelihood', 'NGram', 'GetProbError', 'MolConvertError', 'NGramTrainingError']

from .estimator import GaussianLogLikelihood
from .iqspr import IQSPR
from .modifier import NGram, GetProbError, MolConvertError, NGramTrainingError



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from collections import OrderedDict
from collections import defaultdict
from pathlib import Path
from typing import Any, Union, Dict, Callable, Tuple

import joblib
import pandas as pd
import torch
from deprecated import deprecated
from torch.nn import Module

from xenonpy.model.training.base import BaseRunner

__all__ = ['Checker']


class Checker(object):
    """
    Check point.
    """

    class __SL:
        dump = torch.save
        load = torch.load

    def __init__(self,
                 path: Union[Path, str] = '.',
                 *,
                 increment: bool = False,
                 device: Union[bool, str, torch.device] = 'cpu',
                 default_handle: Tuple[Callable, str] = (joblib, '.pkl.z')):
        """
        Parameters
        ----------
        path: Union[Path, str]
            Dir path for data access. Can be ``Path`` or ``str``.
            Given a relative path will be resolved to abstract path automatically.
        increment : bool
            Set to ``True`` to prevent the potential risk of overwriting.
            Default ``False``.
        """
        if path == '.':
            path = Path(path).resolve()
            path = path / path.name
        else:
            path = Path(path).resolve()
        if increment:
            i = 1
            while Path(f'{path}@{i}').exists():
                i += 1
            path = f'{path}@{i}'
        self._path = Path(path)
        self._path.mkdir(parents=True, exist_ok=True)
        self._device = BaseRunner.check_device(device)
        self._handle = default_handle

        self._files: Dict[str, str] = defaultdict(str)
        self._make_file_index()

    @classmethod
    def load(cls, model_path):
        return cls(model_path)

    @property
    def path(self):
        return str(self._path)

    @property
    def files(self):
        return list(self._files.keys())

    @property
    def model_name(self):
        """

        Returns
        -------
        str
            Model name.
        """
        return self._path.name

    @property
    def model_structure(self):
        structure = self['model_structure']
        print(structure)
        return structure

    @property
    def training_info(self):
        return self['training_info']

    @property
    def describe(self):
        """
        Description for this model.

        Returns
        -------
        dict
            Description.
        """
        return self['describe']

    @property
    def model(self):
        """

        Returns
        -------
        model: :class:`torch.nn.Module`
            A pytorch model.
        """
        if (self._path / 'model.pth.m').exists():
            return torch.load(str(self._path / 'model.pth.m'), map_location=self._device)
        return None

    @model.setter
    def model(self, model: Module):
        """
        Set a model instance.

        Parameters
        ----------
        model: :class:`torch.nn.Module`
            Pytorch model instance.
        """
        if isinstance(model, Module):
            self(model=model)
            self.init_state = model.state_dict()
            self(model_structure=str(model))
        else:
            raise TypeError(f'except `torch.nn.Module` object but got {type(model)}')

    @property
    @deprecated('This property is rotten and will be removed in v0.5.0')
    def trained_model(self):
        if (self._path / 'trained_model.@1.pkl.z').exists():
            return torch.load(str(self._path / 'trained_model.@1.pkl.z'), map_location=self._device)
        else:
            tmp = self.final_state
            if tmp is not None:
                return tmp
        return None

    @property
    def init_state(self):
        if (self._path / 'init_state.pth.s').exists():
            return torch.load(str(self._path / 'init_state.pth.s'), map_location=self._device)
        return None

    @init_state.setter
    def init_state(self, state: OrderedDict):
        if not isinstance(state, OrderedDict) or not state:
            raise TypeError
        for v in state.values():
            if not isinstance(v, torch.Tensor):
                raise TypeError()
        self((Checker.__SL, '.pth.s'), init_state=state)

    @property
    def final_state(self):
        if (self._path / 'final_state.pth.s').exists():
            return torch.load(str(self._path / 'final_state.pth.s'), map_location=self._device)
        return None

    @final_state.setter
    def final_state(self, state: OrderedDict):
        if not isinstance(state, OrderedDict) or not state:
            raise TypeError
        for v in state.values():
            if not isinstance(v, torch.Tensor):
                raise TypeError()
        self((Checker.__SL, '.pth.s'), final_state=state)

    def _make_file_index(self):

        for f in [f for f in self._path.iterdir() if f.match('*.pkl.*') or f.match('*.pd.*') or f.match('*.pth.*')]:
            # select data
            fn = '.'.join(f.name.split('.')[:-2])
            self._files[fn] = str(f)

    def _save_data(self, data: Any, filename: str, handle) -> str:
        if isinstance(data, pd.DataFrame):
            file = str(self._path / (filename + '.pd.xz'))
            self._files[filename] = file
            pd.to_pickle(data, file)
        elif isinstance(data, (torch.Tensor, torch.nn.Module)):
            file = str(self._path / (filename + '.pth.m'))
            self._files[filename] = file
            torch.save(data, file)
        else:
            file = str(self._path / (filename + handle[1]))
            self._files[filename] = file
            handle[0].dump(data, file)

        return file

    def _load_data(self, file: str, handle):
        fp = self._files[file]
        if fp == '':
            return None
        fp_ = Path(fp)
        if not fp_.exists():
            return None
        patten = fp_.name.split('.')[-2]
        if patten == 'pd':
            return pd.read_pickle(fp)
        if patten == 'pth':
            return torch.load(fp, map_location=self._device)
        if patten == 'pkl':
            return joblib.load(fp)
        else:
            return handle.load(fp)

    def __getattr__(self, name: str):
        """
        Return sub-dataset.

        Parameters
        ----------
        name: str
            Dataset name.

        Returns
        -------
        self
        """
        if name == 'checkpoints':
            sub_set = self.__class__(self._path / name, increment=False, device=self._device)
            setattr(self, f'{name}', sub_set)
            return sub_set
        raise AttributeError(f'no such attribute named {name}')

    def __getitem__(self, item):

        if isinstance(item, str):
            return self._load_data(item, self._handle[0])
        else:
            raise KeyError(f'{item}')

    def __call__(self, handle=None, **named_data: Any):
        """
        Save data with or without name.
        Data with same name will not be overwritten.

        Parameters
        ----------
        handle: Tuple[Callable, str]
        named_data: dict
            Named data as k,v pair.

        """
        if handle is None:
            handle = self._handle

        for k, v in named_data.items():
            self._save_data(v, k, handle)

    def set_checkpoint(self, **kwargs):
        self.checkpoints((Checker.__SL, '.pth.s'), **kwargs)

    def __repr__(self):
        cont_ls = ['<{}> includes:'.format(self.__class__.__name__)]

        for k, v in self._files.items():
            cont_ls.append('"{}": {}'.format(k, v))

        return '\n'.join(cont_ls)



#  Copyright (c) 2019. stewu5. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from .frozen_featurizer_descriptor import FrozenFeaturizerDescriptor
from .mordred_descriptor import Mordred2DDescriptor



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from typing import Union, Tuple, Iterable, List

import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from sklearn import utils
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split, KFold

__all__ = ['Splitter']


class Splitter(BaseEstimator):
    """
    Data splitter for train and test
    """

    def __init__(self,
                 size: int,
                 *,
                 test_size: Union[float, int] = 0.2,
                 k_fold: Union[int, Iterable, None] = None,
                 random_state: Union[int, None] = None,
                 shuffle: bool = True):
        """
        Parameters
        ----------
        size
            Total sample size.
            All data must have same length of their first dim,
        test_size
            If float, should be between ``0.0`` and ``1.0`` and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. Can be ``0`` if cv is ``None``.
            In this case, :meth:`~Splitter.cv` will yield a tuple only contains ``training`` and ``validation``
            on each step. By default, the value is set to 0.2.
        k_fold
            Number of k-folds.
            If ``int``, Must be at least 2.
            If ``Iterable``, it should provide label for each element which will be used for group cv.
            In this case, the input of :meth:`~Splitter.cv` must be a :class:`pandas.DataFrame` object.
            Default value is None to specify no cv.
        random_state
            If int, random_state is the seed used by the random number generator;
            Default is None.
        shuffle
            Whether or not to shuffle the data before splitting.
        """
        if k_fold is None and test_size == 0:
            raise RuntimeError('<test_size> can be none only if <cv> is not none')
        self._k_fold = k_fold
        self._shuffle = shuffle
        self._test_size = test_size

        self._sample_size = np.arange(size)
        self._test: Union[np.ndarray, None] = None
        self._train: Union[np.ndarray, None] = None
        self._cv_indices: List[Tuple[np.ndarray, np.ndarray]] = []
        self.roll(random_state)

    @property
    def size(self):
        return self._sample_size.size

    @property
    def shuffle(self):
        return self._shuffle

    @property
    def test_size(self):
        return self._test_size

    def roll(self, random_state: int = None):

        if self._test_size == 0:
            if self._shuffle:
                self._train = utils.shuffle(self._sample_size)
            else:
                self._train = self._sample_size
        else:
            self._train, self._test = train_test_split(
                self._sample_size,
                test_size=self._test_size,
                random_state=random_state,
                shuffle=self._shuffle)

        if isinstance(self._k_fold, int):
            cv = KFold(n_splits=self._k_fold, shuffle=self._shuffle, random_state=random_state)
            for train, val in cv.split(self._train):
                self._cv_indices.append((self._train[train], self._train[val]))
        elif isinstance(self._k_fold, Iterable):
            tmp: pd.Series = pd.Series(self._k_fold).reset_index(drop=True).iloc[self._train]
            for g in set(tmp):
                val = tmp[tmp == g].index.values
                train = tmp[tmp != g].index.values
                self._cv_indices.append((train, val))

    def _size_check(self, array):
        if not isinstance(array, (np.ndarray, pd.DataFrame, pd.Series)):
            raise TypeError(
                f'<arrays> must be numpy.ndarray, pandas.DataFrame, or pandas.Series but got {array.__class__}.')
        if array.shape[0] != self.size:
            raise ValueError(
                f'parameters <arrays> must have size {self.size} for dim 0 but got {array.shape[0]}')

    @staticmethod
    def _split(array, *idx):

        # all to np.array
        if isinstance(array, np.ndarray):
            return [array[i] for i in idx]

        if isinstance(array, (DataFrame, Series)):
            return [array.iloc[i] for i in idx]

    def cv(self, *arrays, less_for_train=False):
        """
        Split data with cross-validation.

        Parameters
        ----------
        *arrays: DataFrame, Series, ndarray, list
            Data for split. Must be a Sequence of indexables with same length / shape[0].
            If None, return the split indices.
        less_for_train: bool
            If true, use less data set for train.
            E.g. ``[1, 2, 3, 4, 5, 6, 7, 8, 9, 0]`` with 5 cv will be split into
            ``[1, 2]`` and ``[3, 4, 5, 6, 7, 8, 9, 0]``. Usually, ``[1, 2]`` (less one)
            will be used for val. With ``less_for_train=True``, ``[1, 2]`` will be
            used for train. Default is ``False``.

        Yields
        -------
        tuple
            list containing split of inputs with cv. if inputs are None, only return
            the indices of split. if ``test_size`` is 0, test data/index will
            not return.
        """

        if self._k_fold is None:
            raise RuntimeError('parameter <cv> must be set')

        for train, val in self._cv_indices:
            if less_for_train:
                tmp = train
                train = val
                val = tmp

            if len(arrays) == 0:
                if self._test is not None:
                    yield train, val, self._test
                else:
                    yield train, val
            else:
                ret = []
                for array in arrays:
                    self._size_check(array)
                    if self._test is not None:
                        ret.extend(self._split(array, train, val, self._test))
                    else:
                        ret.extend(self._split(array, train, val))
                yield tuple(ret)
        return

    def split(self, *arrays: Union[np.ndarray, pd.DataFrame, pd.Series]):
        """
        Split data.

        Parameters
        ----------
        *arrays
            Dataset for split.
            Size of dim 0 must be equal to :meth:`~Splitter.size`.
            If None, return the split indices.

        Returns
        -------
        tuple
            List containing split of inputs. if inputs are None, only return
            the indices of split. if ``test_size`` is 0, test data/index will
            not return.
        """
        if self._test is None:
            raise RuntimeError('split action is illegal because `test_size` is none')

        if len(arrays) == 0:
            return self._train, self._test

        ret = []
        for array in arrays:
            self._size_check(array)
            ret.extend(self._split(array, self._train, self._test))
        return tuple(ret)



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from torch import optim

from xenonpy.model.training.base import BaseOptimizer

__all__ = ['Adadelta', 'Adagrad', 'Adam', 'Adamax', 'ASGD', 'SGD', 'SparseAdam', 'RMSprop', 'Rprop', 'LBFGS']


class Adadelta(BaseOptimizer):

    def __init__(self, *, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0):
        """Implements Adadelta algorithm.

        It has been proposed in `ADADELTA: An Adaptive Learning Rate Method`__.

        Arguments:
            rho (float, optional): coefficient used for computing a running average
                of squared gradients (default: 0.9)
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-6)
            lr (float, optional): coefficient that scale delta before it is applied
                to the parameters (default: 1.0)
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

        __ https://arxiv.org/abs/1212.5701
        """
        super().__init__(optim.Adadelta, lr=lr, rho=rho, eps=eps, weight_decay=weight_decay)


class Adagrad(BaseOptimizer):

    def __init__(self, *, lr=0.01, lr_decay=0, weight_decay=0, initial_accumulator_value=0):
        """Implements Adagrad algorithm.

        It has been proposed in `Adaptive Subgradient Methods for Online Learning
        and Stochastic Optimization`_.

        Arguments:
            lr (float, optional): learning rate (default: 1e-2)
            lr_decay (float, optional): learning rate decay (default: 0)
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

        .. _Adaptive Subgradient Methods for Online Learning and Stochastic
            Optimization: http://jmlr.org/papers/v12/duchi11a.html
        """
        super().__init__(optim.Adagrad, lr=lr, lr_decay=lr_decay, weight_decay=weight_decay,
                         initial_accumulator_value=initial_accumulator_value)


class Adam(BaseOptimizer):

    def __init__(self, *, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, amsgrad=False):
        r"""Implements Adam algorithm.

        It has been proposed in `Adam: A Method for Stochastic Optimization`_.

        Arguments:
            lr (float, optional): learning rate (default: 1e-3)
            betas (Tuple[float, float], optional): coefficients used for computing
                running averages of gradient and its square (default: (0.9, 0.999))
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-8)
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
            amsgrad (boolean, optional): whether to use the AMSGrad variant of this
                algorithm from the paper `On the Convergence of Adam and Beyond`_
                (default: False)

        .. _Adam\: A Method for Stochastic Optimization:
            https://arxiv.org/abs/1412.6980
        .. _On the Convergence of Adam and Beyond:
            https://openreview.net/forum?id=ryQu7f-RZ
        """

        super().__init__(optim.Adam, lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad)


class SparseAdam(BaseOptimizer):

    def __init__(self, *, lr=0.001, betas=(0.9, 0.999), eps=1e-08):
        r"""Implements lazy version of Adam algorithm suitable for sparse tensors.

        In this variant, only moments that show up in the gradient get updated, and
        only those portions of the gradient get applied to the parameters.

        Arguments:
            lr (float, optional): learning rate (default: 1e-3)
            betas (Tuple[float, float], optional): coefficients used for computing
                running averages of gradient and its square (default: (0.9, 0.999))
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-8)

        .. _Adam\: A Method for Stochastic Optimization:
            https://arxiv.org/abs/1412.6980
        """

        super().__init__(optim.SparseAdam, lr=lr, betas=betas, eps=eps)


class Adamax(BaseOptimizer):

    def __init__(self, *, lr=0.002, betas=(0.9, 0.999), eps=1e-08, weight_decay=0):
        """Implements Adamax algorithm (a variant of Adam based on infinity norm).

        It has been proposed in `Adam: A Method for Stochastic Optimization`__.

        Arguments:
            lr (float, optional): learning rate (default: 2e-3)
            betas (Tuple[float, float], optional): coefficients used for computing
                running averages of gradient and its square
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-8)
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

        __ https://arxiv.org/abs/1412.6980
        """

        super().__init__(optim.Adamax, lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)


class ASGD(BaseOptimizer):

    def __init__(self, *, lr=0.002, lambd=0.0001, alpha=0.75, t0=1000000.0, weight_decay=0):
        """Implements Averaged Stochastic Gradient Descent.

        It has been proposed in `Acceleration of stochastic approximation by
        averaging`_.

        Arguments:
            lr (float, optional): learning rate (default: 1e-2)
            lambd (float, optional): decay term (default: 1e-4)
            alpha (float, optional): power for eta update (default: 0.75)
            t0 (float, optional): point at which to start averaging (default: 1e6)
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

        .. _Acceleration of stochastic approximation by averaging:
            http://dl.acm.org/citation.cfm?id=131098
        """

        super().__init__(optim.ASGD, lr=lr, lambd=lambd, alpha=alpha, t0=t0, weight_decay=weight_decay)


class LBFGS(BaseOptimizer):

    def __init__(self, *, lr=1, max_iter=20, max_eval=None,
                 tolerance_grad=1e-5, tolerance_change=1e-9, history_size=100,
                 line_search_fn=None):
        """Implements L-BFGS algorithm.

        .. warning::
            This optimizer doesn't support per-parameter options and parameter
            groups (there can be only one).

        .. warning::
            Right now all parameters have to be on a single device. This will be
            improved in the future.

        .. note::
            This is a very memory intensive optimizer (it requires additional
            ``param_bytes * (history_size + 1)`` bytes). If it doesn't fit in memory
            try reducing the history size, or use a different algorithm.

        Arguments:
            lr (float): learning rate (default: 1)
            max_iter (int): maximal number of iterations per optimization step
                (default: 20)
            max_eval (int): maximal number of function evaluations per optimization
                step (default: max_iter * 1.25).
            tolerance_grad (float): termination tolerance on first order optimality
                (default: 1e-5).
            tolerance_change (float): termination tolerance on function
                value/parameter changes (default: 1e-9).
            history_size (int): update history size (default: 100).
        """

        super().__init__(optim.LBFGS, lr=lr, max_iter=max_iter, max_eval=max_eval,
                         tolerance_grad=tolerance_grad, tolerance_change=tolerance_change, history_size=history_size,
                         line_search_fn=line_search_fn)


class RMSprop(BaseOptimizer):

    def __init__(self, *, lr=0.01, alpha=0.99, eps=1e-08, weight_decay=0, momentum=0, centered=False):
        """Implements RMSprop algorithm.

        Proposed by G. Hinton in his
        `course <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.

        The centered version first appears in `Generating Sequences
        With Recurrent Neural Networks <https://arxiv.org/pdf/1308.0850v5.pdf>`_.

        Arguments:
            lr (float, optional): learning rate (default: 1e-2)
            momentum (float, optional): momentum factor (default: 0)
            alpha (float, optional): smoothing constant (default: 0.99)
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-8)
            centered (bool, optional) : if ``True``, compute the centered RMSProp,
                the gradient is normalized by an estimation of its variance
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

        """

        super().__init__(optim.RMSprop, lr=lr, alpha=alpha, eps=eps, weight_decay=weight_decay, momentum=momentum,
                         centered=centered)


class Rprop(BaseOptimizer):

    def __init__(self, *, lr=0.01, etas=(0.5, 1.2), step_sizes=(1e-06, 50)):
        """Implements the resilient backpropagation algorithm.

        Arguments:
            lr (float, optional): learning rate (default: 1e-2)
            etas (Tuple[float, float], optional): pair of (etaminus, etaplis), that
                are multiplicative increase and decrease factors
                (default: (0.5, 1.2))
            step_sizes (Tuple[float, float], optional): a pair of minimal and
                maximal allowed step sizes (default: (1e-6, 50))
        """

        super().__init__(optim.Rprop, lr=lr, etas=etas, step_sizes=step_sizes)


class SGD(BaseOptimizer):

    def __init__(self, *, lr=0.01, momentum=0, dampening=0, weight_decay=0, nesterov=False):
        r"""Implements stochastic gradient descent (optionally with momentum).

        Nesterov momentum is based on the formula from
        `On the importance of initialization and momentum in deep learning`__.

        Args:
            lr (float): learning rate
            momentum (float, optional): momentum factor (default: 0)
            weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
            dampening (float, optional): dampening for momentum (default: 0)
            nesterov (bool, optional): enables Nesterov momentum (default: False)

        Example:
            >>> optimizer = torch.optim.SGD(lr=0.1, momentum=0.9)
            >>> optimizer(model.parameters())
            >>> optimizer.zero_grad()
            >>> loss_fn(model(input), target).backward()
            >>> optimizer.step(),

        __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf

        .. note::
            The implementation of SGD with Momentum/Nesterov subtly differs from
            Sutskever et. al. and implementations in some other frameworks.

            Considering the specific case of Momentum, the update can be written as

            .. math::
                      v = \rho * v + g \\
                      p = p - lr * v

            where p, g, v and :math:`\rho` denote the parameters, gradient,
            velocity, and momentum respectively.

            This is in contrast to Sutskever et. al. and
            other frameworks which employ an update of the form

            .. math::
                 v = \rho * v + lr * g \\
                 p = p - v

            The Nesterov version is analogously modified.
        """

        super().__init__(optim.SGD, lr=lr, momentum=momentum, dampening=dampening, weight_decay=weight_decay,
                         nesterov=nesterov)



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import torch
from torch import nn

__all__ = ['ConvLayer', 'CrystalGraphConvNet']


class ConvLayer(nn.Module):
    """
    Convolutional operation on graphs
    """

    def __init__(self, atom_fea_len, nbr_fea_len):
        """
        Initialize ConvLayer.

        Parameters
        ----------

        atom_fea_len: int
          Number of atom hidden features.
        nbr_fea_len: int
          Number of bond features.
        """
        super(ConvLayer, self).__init__()
        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        self.fc_full = nn.Linear(2 * self.atom_fea_len + self.nbr_fea_len,
                                 2 * self.atom_fea_len)
        self.sigmoid = nn.Sigmoid()
        self.softplus1 = nn.Softplus()
        self.bn1 = nn.BatchNorm1d(2 * self.atom_fea_len)
        self.bn2 = nn.BatchNorm1d(self.atom_fea_len)
        self.softplus2 = nn.Softplus()

    def forward(self, atom_in_fea, nbr_fea, nbr_fea_idx):
        """
        Forward pass

        N: Total number of atoms in the batch
        M: Max number of neighbors

        Parameters
        ----------

        atom_in_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
          Atom hidden features before convolution
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors
        nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom

        Returns
        -------

        atom_out_fea: nn.Variable shape (N, atom_fea_len)
          Atom hidden features after convolution

        """
        # TODO will there be problems with the index zero padding?
        N, M = nbr_fea_idx.shape
        # convolution
        atom_nbr_fea = atom_in_fea[nbr_fea_idx, :]
        total_nbr_fea = torch.cat(
            [atom_in_fea.unsqueeze(1).expand(N, M, self.atom_fea_len),
             atom_nbr_fea, nbr_fea], dim=2)
        total_gated_fea = self.fc_full(total_nbr_fea)
        total_gated_fea = self.bn1(total_gated_fea.view(
            -1, self.atom_fea_len * 2)).view(N, M, self.atom_fea_len * 2)
        nbr_filter, nbr_core = total_gated_fea.chunk(2, dim=2)
        nbr_filter = self.sigmoid(nbr_filter)
        nbr_core = self.softplus1(nbr_core)
        nbr_sumed = torch.sum(nbr_filter * nbr_core, dim=1)
        nbr_sumed = self.bn2(nbr_sumed)
        out = self.softplus2(atom_in_fea + nbr_sumed)
        return out


class CrystalGraphConvNet(nn.Module):
    """
    Create a crystal graph convolutional neural network for predicting total
    material properties.

    See Also: [CGCNN]_.

    .. [CGCNN] `Crystal Graph Convolutional Neural Networks for an Accurate and Interpretable Prediction of Material Properties`__
    __ https://doi.org/10.1103/PhysRevLett.120.145301
    """

    def __init__(self, orig_atom_fea_len, nbr_fea_len,
                 atom_fea_len=64, n_conv=3, h_fea_len=128, n_h=1,
                 classification=False):
        """
        Initialize CrystalGraphConvNet.

        Parameters
        ----------

        orig_atom_fea_len: int
          Number of atom features in the input.
        nbr_fea_len: int
          Number of bond features.
        atom_fea_len: int
          Number of hidden atom features in the convolutional layers
        n_conv: int
          Number of convolutional layers
        h_fea_len: int
          Number of hidden features after pooling
        n_h: int
          Number of hidden layers after pooling
        """
        super(CrystalGraphConvNet, self).__init__()
        self.classification = classification
        self.embedding = nn.Linear(orig_atom_fea_len, atom_fea_len)
        self.convs = nn.ModuleList([ConvLayer(atom_fea_len=atom_fea_len,
                                              nbr_fea_len=nbr_fea_len)
                                    for _ in range(n_conv)])
        self.conv_to_fc = nn.Linear(atom_fea_len, h_fea_len)
        self.conv_to_fc_softplus = nn.Softplus()
        if n_h > 1:
            self.fcs = nn.ModuleList([nn.Linear(h_fea_len, h_fea_len)
                                      for _ in range(n_h - 1)])
            self.softpluses = nn.ModuleList([nn.Softplus()
                                             for _ in range(n_h - 1)])
        if self.classification:
            self.fc_out = nn.Linear(h_fea_len, 2)
        else:
            self.fc_out = nn.Linear(h_fea_len, 1)
        if self.classification:
            self.logsoftmax = nn.LogSoftmax()
            self.dropout = nn.Dropout()

    def forward(self, atom_fea, nbr_fea, nbr_fea_idx, crystal_atom_idx):
        """
        Forward pass

        N: Total number of atoms in the batch
        M: Max number of neighbors
        N0: Total number of crystals in the batch

        Parameters
        ----------

        atom_fea: Variable(torch.Tensor) shape (N, orig_atom_fea_len)
          Atom features from atom type
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors
        nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom
        crystal_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx

        Returns
        -------

        prediction: nn.Variable shape (N, )
          Atom hidden features after convolution

        """
        atom_fea = self.embedding(atom_fea)
        for conv_func in self.convs:
            atom_fea = conv_func(atom_fea, nbr_fea, nbr_fea_idx)
        crys_fea = self.pooling(atom_fea, crystal_atom_idx)
        crys_fea = self.conv_to_fc(self.conv_to_fc_softplus(crys_fea))
        crys_fea = self.conv_to_fc_softplus(crys_fea)
        if self.classification:
            crys_fea = self.dropout(crys_fea)
        if hasattr(self, 'fcs') and hasattr(self, 'softpluses'):
            for fc, softplus in zip(self.fcs, self.softpluses):
                crys_fea = softplus(fc(crys_fea))
        out = self.fc_out(crys_fea)
        if self.classification:
            out = self.logsoftmax(out)
        return out

    @staticmethod
    def pooling(atom_fea, crystal_atom_idx):
        """
        Pooling the atom features to crystal features

        N: Total number of atoms in the batch
        N0: Total number of crystals in the batch

        Parameters
        ----------

        atom_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
          Atom feature vectors of the batch
        crystal_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx
        """
        assert sum([len(idx_map) for idx_map in crystal_atom_idx]) == \
               atom_fea.data.shape[0]
        summed_fea = [torch.mean(atom_fea[idx_map], dim=0, keepdim=True)
                      for idx_map in crystal_atom_idx]
        return torch.cat(summed_fea, dim=0)



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from copy import deepcopy
from types import MethodType
from typing import Union

import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.base import BaseEstimator
from sklearn.linear_model import BayesianRidge

from xenonpy.descriptor.base import BaseDescriptor, BaseFeaturizer
from xenonpy.inverse.base import BaseLogLikelihood


class GaussianLogLikelihood(BaseLogLikelihood):
    def __init__(self, descriptor: Union[BaseFeaturizer, BaseDescriptor], *, targets={}, **estimators: BaseEstimator):
        """
        Gaussian loglikelihood.
        Parameters
        ----------
        descriptor: BaseFeaturizer or BaseDescriptor
            Descriptor calculator.
        estimators: BaseEstimator
            Gaussian estimators follow the scikit-learn style.
            These estimators must provide a method named ``predict`` which
            accesses descriptors as input and returns ``(mean, std)`` in order.
            By default, BayesianRidge_ will be used.
            .. _BayesianRidge: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html#sklearn-linear-model-bayesianridge
        targets: dictionary
            Upper and lower bounds for each property to calculate the Gaussian CDF probability
        """
        if estimators:
            self._mdl = deepcopy(estimators)
        else:
            self._mdl = {}

        if not isinstance(descriptor, (BaseFeaturizer, BaseDescriptor)):
            raise TypeError('<descriptor> must be a subclass of <BaseFeaturizer> or <BaseDescriptor>')
        self._descriptor = descriptor
        self._descriptor.set_params(on_errors='nan')

        self._targets = deepcopy(targets)

    def __getitem__(self, item):
        return self._mdl[item]

    def __setitem__(self, key, value):
        if not (hasattr(value, 'predict') and isinstance(value.predict, MethodType)):
            raise TypeError('estimator must be a regressor in scikit-learn style')
        self._mdl[key] = deepcopy(value)

    def update_targets(self, *, reset=False, **targets):
        """
        Update/set the target area.
        Parameters
        ----------
        reset: bool
            If ``true``, reset target area.
        targets: tuple[float, float]
            Target area.
            Should be a tuple which have down and up boundary.
            e.g: ``target1=(10, 20)`` equal to ``target1 should in range [10, 20]``.
        """
        if reset:
            self._targets = {}
        for k, v in targets.items():
            if not isinstance(v, tuple) or len(v) != 2 or v[1] <= v[0]:
                raise ValueError('must be a tuple with (low, up) boundary')
            self._targets[k] = v

    def remove_estimator(self, *properties: str):
        """
        Remove estimators from estimator set.
        Parameters
        ----------
        properties : str
            The name of properties will be removed from estimator set.
        """
        if not properties:
            self._mdl = {}
            self._targets = {}
        else:
            for p in properties:
                del self._mdl[p]
                del self._targets[p]


    def predict(self, smiles, **kwargs):
        fps = self._descriptor.transform(smiles, return_type='df')
        fps_ = fps.dropna()
        tmp = {}
        for k, v in self._mdl.items():
            if isinstance(v, BayesianRidge):
                tmp[k + ': mean'], tmp[k + ': std'] = v.predict(fps_, return_std=True)
            else:
                tmp[k + ': mean'], tmp[k + ': std'] = v.predict(fps_, **kwargs)

        tmp = pd.DataFrame(data=tmp, index=fps_.index)
        return pd.DataFrame(data=tmp, index=fps.index)

    # todo: implement scale function
    def fit(self, smiles, y=None, *, X_scaler=None, y_scaler=None, **kwargs):
        """
        Default - automatically remove NaN data rows
        
        Parameters
        ----------
        smiles: list[str]
            SMILES for training.
        y: pandas.DataFrame
            Target properties for training.
        X_scaler: Scaler (optional, not implement)
            Scaler for transform X.
        y_scaler: Scaler (optional, not implement)
            Scaler for transform y.
        kwargs: dict
            Parameters pass to BayesianRidge initialization.
        """

        if self._mdl:
            raise RuntimeError('estimators have been set.'
                               'If you want to re-train these estimators,'
                               'please use `remove_estimator()` method first.')

        if not isinstance(y, (pd.DataFrame, pd.Series)):
            raise TypeError('please package all properties into a pd.DataFrame or pd.Series')

        # remove NaN from X
        desc = self._descriptor.transform(smiles, return_type='df').reset_index(drop=True)
        y = y.reset_index(drop=True)
        desc.dropna(inplace=True)
        y = pd.DataFrame(y.loc[desc.index])

        for c in y:
            y_ = y[c]  # get target property.
            # remove NaN from y_
            y_.dropna(inplace=True)
            desc_ = desc.loc[y_.index]
            desc_ = desc_.values

            mdl = BayesianRidge(compute_score=True, **kwargs)
            mdl.fit(desc_, y_)
            self._mdl[c] = mdl

    # log_likelihood returns a dataframe of log-likelihood values of each property & sample
    def log_likelihood(self, smis, **targets):
        def _avoid_overflow(ll_):
            # log(exp(log(UP) - log(C)) - exp(log(LOW) - log(C))) + log(C)
            # where C = max(log(UP), max(LOW))
            ll_c = np.max(ll_)
            ll_ = np.log(np.exp(ll_[1] - ll_c) - np.exp(ll_[0] - ll_c)) + ll_c
            return ll_

        # self.update_targets(reset=False, **targets):
        for k, v in targets.items():
            if not isinstance(v, tuple) or len(v) != 2 or v[1] <= v[0]:
                raise ValueError('must be a tuple with (low, up) boundary')
            self._targets[k] = v

        if not self._targets:
            raise RuntimeError('<targets> is empty')

        ll = pd.DataFrame(np.full((len(smis), len(self._mdl)), -1000.0), columns=self._mdl.keys())
        pred = self.predict(smis).reset_index(drop=True)
        tmp = pred.isna().any(axis=1)
        idx = [i for i in range(len(smis)) if ~tmp[i]]

        # calculate likelihood
        for k, (low, up) in self._targets.items():  # k: target; v: (low, up)

            # predict mean, std for all smiles
            mean, std = pred[k + ': mean'], pred[k + ': std']

            # calculate low likelihood
            low_ll = norm.logcdf(low, loc=np.asarray(mean), scale=np.asarray(std))

            # calculate up likelihood
            up_ll = norm.logcdf(up, loc=np.asarray(mean), scale=np.asarray(std))

            # zip low and up likelihood to a 1-dim array then save it.
            # like: [(tar_low_smi1, tar_up_smi1),  (tar_low_smi2, tar_up_smi2), ..., (tar_low_smiN, tar_up_smiN)]
            lls = zip(low_ll, up_ll)
            ll[k].iloc[idx] = np.array([*map(_avoid_overflow, list(lls))])

        return ll
